from attr import dataclass
from typing import Optional


@dataclass
class BridgeVlan:
    bridge: str
    current_tagged: str
    current_untagged: str
    disabled: bool
    dynamic: bool
    tagged: str
    untagged: str
    vlan_ids: int
    id: str
    comment: Optional[str] = None



from attr import dataclass
from typing import Optional

from ros._literals import IPProtocol, MACProtocol


@dataclass(slots=True)
class Torch:
    section: int
    rx: int
    rx_packets: int
    tx: int
    tx_packets: int
    cpu: Optional[int] = None
    dscp: Optional[int] = None
    src_port: Optional[str] = None
    dst_port: Optional[str] = None
    ip_protocol: Optional[IPProtocol] = None
    mac_protocol: Optional[MACProtocol] = None
    src_address: Optional[str] = None
    dst_address: Optional[str] = None



from attr import dataclass
from typing import Optional


@dataclass(slots=True)
class Error(Exception):
    error: int
    message: str
    detail: Optional[str] = None

    def __str__(self) -> str:
        if self.detail:
            return f"{self.message} ({self.error}): {self.detail}"
        return f"{self.message} ({self.error})"



from attr import dataclass


@dataclass(slots=True)
class RouterBOARD:
    current_firmware: str
    factory_firmware: str
    firmware_type: str
    model: str
    routerboard: bool
    serial_number: str
    upgrade_firmware: str



from attr import dataclass, field
from typing import Literal, Optional


@dataclass
class Route:
    dst_address: str
    gateway: str
    routing_table: str
    check_gateway: Literal["arp", "bfd", "bfd-multihop", "none", "ping"] = None
    suppress_hw_offload: bool = False
    distance: int = 1
    scope: int = 30
    target_scope: Optional[int] = None
    vrf_interface: Optional[str] = None
    pref_src: Optional[str] = None
    vpn: Optional[bool] = None
    comment: Optional[str] = None
    immediate_gw: str = field(on_setattr=None, default=None)
    active: bool = field(on_setattr=None, default=None)
    dynamic: bool = field(on_setattr=None, default=None)
    ecmp: bool = field(on_setattr=None, default=None)
    hw_offloaded: bool = field(on_setattr=None, default=None)
    inactive: bool = field(on_setattr=None, default=None)
    static: bool = field(on_setattr=None, default=None)
    filtered: bool = field(on_setattr=None, default=None)
    id: str = field(on_setattr=None, default=None)



from typing import List

from ros._base import BaseModule, BaseProps

from .aaa import PPPAAA
from .profile import PPPProfile
from .secret import PPPSecret


class PPPModule(BaseModule):
    _profile: BaseProps[PPPProfile] = None
    _secret: BaseProps[PPPSecret] = None

    @property
    def aaa(self) -> PPPAAA:
        return self.ros.get_as(self.url + "/aaa", PPPAAA)

    @property
    def profile(self):
        if not self._profile:
            self._profile = BaseProps(self, self.url + "/profile", PPPProfile)
        return self._profile

    @property
    def secret(self):
        if not self._secret:
            self._secret = BaseProps(self, self.url + "/secret", PPPSecret)
        return self._secret


__all__ = ["PPPAAA", "PPPModule", "PPPProfile", "PPPSecret"]



from attr import dataclass, field
from typing import Literal, Optional

AddDefaultRoute = Literal["yes", "no", "special-classless"]
Status = Literal[
    "bound", "error", "rebinding...", "requesting...", "searching...", "stopped"
]


@dataclass
class DHCPClient:
    interface: str
    add_default_route: AddDefaultRoute = "yes"
    dhcp_options: str = "hostname,clientid"
    disabled: bool = False
    use_peer_dns: bool = True
    use_peer_ntp: bool = True
    default_route_distance: int = 1
    comment: Optional[str] = None
    address: Optional[str] = None
    dhcp_server: Optional[str] = field(on_setattr=None, default=None)
    expires_after: Optional[str] = field(on_setattr=None, default=None)
    gateway: Optional[str] = field(on_setattr=None, default=None)
    primary_dns: Optional[str] = field(on_setattr=None, default=None)
    primary_ntp: Optional[str] = field(on_setattr=None, default=None)
    secondary_dns: Optional[str] = field(on_setattr=None, default=None)
    status: Status = field(on_setattr=None, default=None)
    dynamic: bool = field(on_setattr=None, default=None)
    invalid: bool = field(on_setattr=None, default=None)
    id: str = field(on_setattr=None, default=None)

    def __str__(self) -> str:
        return self.interface



from attr import dataclass
from typing import Optional


@dataclass(slots=True)
class Resource:
    architecture_name: str
    board_name: str
    build_time: str
    cpu: str
    cpu_count: int
    cpu_load: int
    free_hdd_space: int
    free_memory: int
    platform: str
    total_hdd_space: int
    total_memory: int
    uptime: str
    version: str
    bad_blocks: Optional[float] = None
    cpu_frequency: Optional[int] = None
    write_sect_since_reboot: Optional[int] = None
    write_sect_total: Optional[int] = None



from typing import Optional

from ros._base import BaseModule, BaseProps

from .address import Address
from .arp import ARP
from .cloud import Cloud
from .dhcp_client import DHCPClient
from .dhcp_relay import DHCPRelay
from .dhcp_server import DHCPServerModule, DHCPServer
from .dns import DNS
from .firewall import IPFirewallModule
from .route import Route
from .setting import Setting


class IPModule(BaseModule):
    _dhcp_server: Optional[DHCPServerModule] = None
    _firewall: Optional[IPFirewallModule] = None

    _address: BaseProps[Address] = None
    _arp: BaseProps[ARP] = None
    _dhcp_client: BaseProps[DHCPClient] = None
    _dhcp_relay: BaseProps[DHCPRelay] = None
    _route: BaseProps[Route] = None

    def __attrs_post_init__(self) -> None:
        return super().__attrs_post_init__()

    @property
    def address(self):
        if not self._address:
            self._address = BaseProps(self, self.url + "/address", Address)
        return self._address

    @property
    def arp(self):
        if not self._arp:
            self._arp = BaseProps(self, self.url + "/arp", ARP)
        return self._arp

    @property
    def cloud(self) -> Cloud:
        return self.ros.get_as(self.url + "/cloud", Cloud)

    @property
    def dhcp_client(self):
        if not self._dhcp_client:
            self._dhcp_client = BaseProps(self, self.url + "/dhcp-client", DHCPClient)
        return self._dhcp_client

    @property
    def dhcp_relay(self):
        if not self._dhcp_relay:
            self._dhcp_relay = BaseProps(self, self.url + "/dhcp-relay", DHCPRelay)
        return self._dhcp_relay

    @property
    def dhcp_server(self) -> DHCPServerModule:
        if not self._dhcp_server:
            self._dhcp_server = DHCPServerModule(
                self, self.url + "/dhcp-server", DHCPServer
            )
        return self._dhcp_server

    @property
    def dns(self) -> DNS:
        dns = self.ros.get_as(self.url + "/dns", DNS)
        dns._mod = self
        return dns

    @property
    def firewall(self) -> IPFirewallModule:
        if not self._firewall:
            self._firewall = IPFirewallModule(self, self.url + "/firewall")
        return self._firewall

    @property
    def route(self):
        if not self._route:
            self._route = BaseProps(self, self.url + "/route", Route)
        return self._route

    @property
    def setting(self) -> Setting:
        return self.ros.get_as(self.url + "/setting", Setting)


__all__ = [
    "Address",
    "ARP",
    "IPModule",
    "Cloud",
    "DHCPClient",
    "DHCPRelay",
    "DHCPServer",
    "DNS",
    "Route",
    "Setting",
]



from attr import dataclass
from typing import Optional, Union


@dataclass
class Interface:
    actual_mtu: int
    disabled: bool
    fp_rx_byte: int
    fp_rx_packet: int
    fp_tx_byte: int
    fp_tx_packet: int
    link_downs: int
    mtu: Union[int, str]
    name: str
    running: bool
    rx_byte: int
    rx_drop: int
    rx_error: int
    rx_packet: int
    tx_byte: int
    tx_drop: int
    tx_error: int
    tx_packet: int
    tx_queue_drop: int
    type: str
    id: str
    comment: Optional[str] = None
    l2mtu: Optional[int] = None
    mac_address: Optional[str] = None
    default_name: Optional[str] = None
    max_l2mtu: Optional[int] = None
    slave: Optional[bool] = None

    def __str__(self) -> str:
        return self.name



from ros._base import BaseModule, BaseProps

from .table import RoutingTable


class RoutingModule(BaseModule):
    _table: BaseProps[RoutingTable] = None

    @property
    def table(self):
        if not self._table:
            self._table = BaseProps(self, self.url + "/table", RoutingTable)
        return self._table


__all__ = ["RoutingModule", "RoutingTable"]



from attr import dataclass
from typing import Optional

from ._literals import YesNoDefault, YesNoDefaultRequired


@dataclass
class PPPProfile:
    id: str
    address_list: str
    bridge_learning: YesNoDefault
    change_tcp_mss: YesNoDefault
    default: bool
    name: str
    on_down: str
    on_up: str
    only_one: YesNoDefault
    use_compression: YesNoDefault
    use_encryption: YesNoDefaultRequired
    use_ipv6: YesNoDefaultRequired
    use_mpls: YesNoDefaultRequired
    use_upnp: YesNoDefault
    bridge: Optional[str] = None
    bridge_horizon: Optional[int] = None
    bridge_path_cost: Optional[int] = None
    bridge_port_priority: Optional[str] = None
    comment: Optional[str] = None
    dhcpv6_pd_pool: Optional[str] = None
    dns_server: Optional[str] = None
    idle_timeout: Optional[str] = None
    incoming_filter: Optional[str] = None
    insert_queue_before: Optional[str] = None
    interface_list: Optional[str] = None
    local_address: Optional[str] = None
    outgoing_filter: Optional[str] = None
    parent_queue: Optional[str] = None
    queue_type: Optional[str] = None
    rate_limit: Optional[str] = None
    remote_address: Optional[str] = None
    remote_ipv6_prefix_pool: Optional[str] = None
    session_timeout: Optional[str] = None
    wins_server: Optional[str] = None



from attr import define, asdict
from cattrs import unstructure
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Type, TypeVar, Union

if TYPE_CHECKING:
    from ros.ros import Ros

from ._utils import clean_before_put


@define
class BaseModule:
    ros: "Ros"
    url: str = ""

    def __attrs_post_init__(self) -> None:
        if not self.url:
            cname = self.__class__.__name__.lower()
            self.url = "/" + cname.replace("module", "")


BM = TypeVar("BM", bound=BaseModule)


PR = TypeVar("PR", bound=object)


@define
class BaseProp(Generic[PR]):
    ros: "Ros"
    url: str
    cl: Type[PR]

    def __call__(self, **kwds: Any) -> PR:
        return self.print(**kwds)

    def print(self, **kwds: Any) -> PR:
        return self.ros.get_as(self.url, self.cl, kwds)

    def set(self, **kwds: Any) -> PR:
        return self.ros.post_as(self.url + "/set", None, kwds)


@define
class BaseProps(Generic[PR]):
    mod: BaseModule
    url: str
    cl: Type[PR]
    _create: bool = True
    _delete: bool = True
    _write: bool = True

    @property
    def ros(self):
        return self.mod.ros

    @property
    def filename(self):
        return self.url

    def __attrs_post_init__(self) -> None:
        if not self.url.startswith(self.mod.url):
            self.url = self.mod.url.rstrip("/") + self.url

    def __call__(self, **kwds: Any) -> List[PR]:
        return self.print(**kwds)

    @staticmethod
    def _getid(o: PR):
        assert hasattr(o, "id")
        return getattr(o, "id")

    def add(self, o: PR) -> PR:
        assert self._write and self._create, "Not writeable"
        data = unstructure(o)
        data = clean_before_put(data)
        return self.mod.ros.put_as(self.url, self.cl, data)

    def delete(self, o: PR):
        assert self._write and self._delete, "Not writeable"
        return self.remove(o)

    def _disabled(self, o: PR, s: bool) -> PR:
        assert self._write, "Not writeable"
        return self.mod.ros.patch_as(
            self.url + f"/{self._getid(o)}", self.cl, {"disabled": s}
        )

    def disable(self, o: PR) -> PR:
        return self._disabled(o, True)

    def enable(self, o: PR) -> PR:
        return self._disabled(o, False)

    def print(self, **kwds: Any) -> List[PR]:
        return self.mod.ros.get_as(self.url, List[self.cl], kwds)

    def remove(self, o: PR):
        assert self._write, "Not writeable"
        self.mod.ros.session.delete(self.url + f"/{self._getid(o)}")

    def set(self, o: PR, nw: Dict[str, Any]):
        assert self._write, "Not writeable"
        return self.mod.ros.patch_as(self.url + f"/{self._getid(o)}", self.cl, nw)

    def unset(self):
        # assert self._write, "Not writeable"
        raise NotImplementedError("/unset function has not been implemented")



from typing import Optional
from attr import dataclass

from ._literals import DNSType


@dataclass
class DNSStatic:
    id: str
    disabled: bool
    ttl: str
    dynamic: bool = False
    address: Optional[str] = None
    cname: Optional[str] = None
    comment: Optional[str] = None
    forward_to: Optional[str] = None
    ipv6_address: Optional[str] = None
    name: Optional[str] = None
    ns: Optional[str] = None
    mx_preference: Optional[str] = None
    mx_exchange: Optional[str] = None
    regexp: Optional[str] = None
    srv_priority: Optional[int] = None
    srv_weight: Optional[int] = None
    srv_port: Optional[int] = None
    srv_target: Optional[str] = None
    text: Optional[str] = None
    type: Optional[DNSType] = None

    def __bool__(self) -> bool:
        return not self.disabled



from typing import Any, List, Optional

from ros._base import BaseModule, BaseProps

from .bridge import BridgeModule, Bridge, BridgeMsti, BridgePort, BridgeVlan
from .ethernet import InterfaceEthernet
from .interface import Interface
from .list import InterfaceList, InterfaceListMember, InterfaceListModule


class InterfaceModule(BaseModule):
    _brigde: Optional[BridgeModule] = None
    _ethernet: Optional[BaseProps[InterfaceEthernet]] = None
    _list: Optional[InterfaceListModule] = None

    def __call__(self, **kwds: Any):
        return self.print(**kwds)

    @property
    def bridge(self) -> BridgeModule:
        if not self._brigde:
            self._brigde = BridgeModule(self, "/interface/bridge", Bridge)
        return self._brigde

    @property
    def list(self) -> InterfaceListModule:
        if not self._list:
            self._list = InterfaceListModule(self, "/list", InterfaceList)
        return self._list

    @property
    def ethernet(self) -> BaseProps[InterfaceEthernet]:
        if not self._ethernet:
            self._ethernet = BaseProps(self, "/ethernet", InterfaceEthernet)
            self._ethernet._create = False
            self._ethernet._delete = False
        return self._ethernet

    def print(self, **kwds: Any) -> List[Interface]:
        return self.ros.get_as(self.url, List[Interface], kwds)


__all__ = [
    "BridgeModule",
    "Bridge",
    "BridgeMsti",
    "BridgePort",
    "BridgeVlan",
    "EthernetModule",
    "InterfaceModule",
    "Interface",
    "InterfaceList",
    "InterfaceListMember",
    "InterfaceListModule",
]



from ros._base import BaseProps

from ._literals import UpdateChannel
from .package import Package
from .update import Update, UpdateModule


class PackageModule(BaseProps[Package]):
    _update: UpdateModule = None

    def downgrade(self):
        self.ros.post_as(self.url + "/downgrade", None)

    @property
    def update(self) -> UpdateModule:
        if not self._update:
            self._update = UpdateModule(self.ros, self.url + "/update", Update)
        return self._update


__all__ = ["Package", "PackageModule", "Update", "UpdateChannel", "UpdateModule"]



from typing import Optional
from attr import dataclass


@dataclass(slots=True)
class Cloud:
    ddns_enabled: bool
    ddns_update_interval: Optional[str]
    update_time: bool
    public_address: Optional[str] = None
    status: Optional[str] = None



from attr import dataclass
from typing import Optional

from ._literals import DNSType


@dataclass(slots=True)
class DNSCache:
    id: str
    name: str
    static: bool
    ttl: str
    type: DNSType
    data: Optional[str] = None



from attr import dataclass
from typing import Optional


@dataclass
class InterfaceListMember:
    interface: str
    list: str
    disabled: bool
    dynamic: bool
    id: str
    comment: Optional[str] = None



from attrs import asdict
from cattrs import Converter
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type, Union, TypeVar

if TYPE_CHECKING:
    from ._base import BM


def clean_key(d: Dict[str, Any]) -> dict:
    nd = dict()
    for k, v in d.items():
        k = k.replace("-", "_")
        k = k.replace(".", "")
        nd[k] = v
    return nd


def clean_data(d: Dict[str, Any]) -> dict:
    data: Any = None
    if isinstance(d, dict):
        data = clean_key(d)
    elif isinstance(d, list):
        data = list()
        for val in d:
            if isinstance(val, dict):
                data.append(clean_key(val))
            else:
                data.append(val)
    return data


def clean_filters(d: Dict[str, Any]) -> dict:
    if not d:
        return d
    nd = dict()
    for k, v in d.items():
        k = k.replace("_", "-")
        nd[k] = v
    return nd


def clean_before_put(d: Dict[str, Any]) -> dict:
    if not d:
        return d
    d.pop("id")
    nd = dict()
    for k, v in d.items():
        if v != None:
            nd[k] = v
    return nd


def _union_str_int(v: str, t: Any) -> Union[str, int]:
    if v.isdigit():
        return int(v)
    return v


def dict_diff(old: Dict[str, Any], new: Dict[str, Any]) -> Dict[str, Any]:
    results: Dict[str, Any] = dict()
    for o, ov in old.items():
        nv = new.get(o)
        if nv is None:
            continue
        if nv == ov:
            results[o] = nv
    return results


T = TypeVar("T", bound=object)


def make_setters(url: str, cls: Type[T]):
    def setters(self: "BM", values: List[T]):
        idattr = "id"
        existed = self.simple
        updated: List[Tuple[T, ...]] = list()
        news: List[T] = list()
        results: List[T] = list()
        for queue in values:
            exist = filter(lambda x: getattr(x, idattr) == getattr(queue, id), existed)
            if exist:
                e: T = set(exist)[0]
                assert e
                if e != queue:
                    updated.append((e, queue))
            else:
                news.append(queue)
        for update in updated:
            old, new = update
            _id = getattr(old, idattr)
            data = asdict(new, True)
            newdata = dict_diff(asdict(old), asdict(new))
            res = self.ros.patch_as(self.url + f"{url}/{_id}", cls, newdata)
            if res:
                results.append(res)
        for new in news:
            data = asdict(new)
            data.pop(idattr)
            res = self.ros.put_as(self.url + f"{url}", cls, data)
            if res:
                results.append(res)
        return results

    return setters


def make_converter() -> Converter:
    c = Converter()
    # Main
    c.register_structure_hook(Union[int, str], _union_str_int)
    return c


def setter(self, value):
    # TODO: Implement setter
    pass


def deleter(self, value):
    # TODO: Implement deleter
    pass



from attr import dataclass
from typing import Optional


@dataclass
class BridgeMsti:
    bridge: str
    disabled: bool
    identifier: int
    priority: str
    vlan_mapping: int
    id: str
    comment: Optional[str] = None



from attr import dataclass, field
from typing import Optional


@dataclass
class ARP:
    address: str
    interface: str
    mac_address: Optional[str] = None
    published: bool = False
    disabled: bool = False
    comment: Optional[str] = None
    id: str = field(on_setattr=None, default=None)
    DHCP: Optional[bool] = field(on_setattr=None, default=None)
    complete: Optional[bool] = field(on_setattr=None, default=None)
    dynamic: Optional[bool] = field(on_setattr=None, default=None)
    invalid: Optional[bool] = field(on_setattr=None, default=None)



from attr import define
from typing import Optional, List

from ros._base import BaseProps

from .list import InterfaceList
from .member import InterfaceListMember


@define
class InterfaceListModule(BaseProps[InterfaceList]):
    _member: Optional[BaseProps[InterfaceListMember]] = None

    @property
    def member(self) -> BaseProps[InterfaceListMember]:
        if not self._member:
            self._member = BaseProps(self.mod, "/list/member", InterfaceListMember)
        return self._member


__all__ = ["InterfaceList", "InterfaceListMember", "InterfaceListModule"]



from attr import dataclass
from typing import Optional


@dataclass
class InterfaceList:
    name: str
    exclude: str
    include: str
    builtin: bool
    dynamic: bool
    id: str = None
    comment: Optional[str] = None

    def __str__(self) -> str:
        return self.name



from attr import dataclass
from typing import Optional


@dataclass(slots=True)
class RoutingTable:
    id: str
    name: str
    invalid: bool
    dynamic: bool = False
    fib: str = "no"
    comment: Optional[str] = None

    def __attrs_post_init__(self):
        if self.fib == "":
            self.fib = "yes"



from attr import dataclass
from typing import Optional


@dataclass
class Address:
    address: str
    interface: str
    network: Optional[str] = None
    comment: Optional[str] = None
    actual_interface: Optional[str] = None
    disabled: bool = False
    invalid: Optional[str] = None
    dynamic: Optional[bool] = None
    id: str = None



try:
    import ujson as json
except ImportError:
    import json  # type: ignore[no-redef]
from attr import define
from requests import Session
from requests.auth import HTTPBasicAuth
from typing import Any, Dict, List, Optional, Type, TypeVar

from . import (
    InterfaceModule,
    IPModule,
    PPPModule,
    QueueModule,
    RoutingModule,
    SystemModule,
    ToolModule,
)

from . import Error, Log

from .inteface import BridgeModule
from ._utils import clean_data, clean_filters, make_converter

_converter = make_converter()

T = TypeVar("T", bound=object)


@define
class BaseRos:
    server: str
    username: str
    password: str
    session: Session = Session()
    secure: bool = False
    filename: str = "rest"
    url: str = ""

    def __attrs_post_init__(self) -> None:
        if not self.server.endswith("/"):
            self.server += "/"
        self.session.auth = HTTPBasicAuth(self.username, self.password)
        self.session.verify = self.secure
        self.password = ""
        self.url = self.server + self.filename

    def get_as(self, filename: str, cl: Type[T], filters: Dict[str, Any] = None) -> T:
        res = self.session.get(
            self.url + filename,
            params=clean_filters(filters),
            verify=self.secure,
        )
        odata = json.loads(res.text)
        data: Any = clean_data(odata)
        if data and "error" in data:
            raise _converter.structure(data, Error)
        try:
            return _converter.structure(data, cl)
        except Exception as e:
            raise e

    def post_as(
        self,
        filename: str,
        cl: Type[T],
        json_: Any = None,
        data: Any = None,
    ) -> T:
        res = self.session.post(
            self.url + filename,
            data=data,
            json=json_,
        )
        odata = json.loads(res.text)
        data = clean_data(odata)
        if data and "error" in data:
            raise _converter.structure(data, Error)
        return _converter.structure(data, cl)

    def patch_as(
        self,
        filename: str,
        cl: Type[T],
        json_: Any = None,
        data: Any = None,
    ) -> T:
        res = self.session.patch(
            self.url + filename,
            data=data,
            json=json_,
        )
        odata = json.loads(res.text)
        data = clean_data(odata)
        if data and "error" in data:
            raise _converter.structure(data, Error)
        return _converter.structure(data, cl)

    def put_as(
        self,
        filename: str,
        cl: Type[T],
        json_: Any = None,
        data: Any = None,
    ) -> T:
        res = self.session.put(
            self.url + filename,
            data=data,
            json=json_,
        )
        odata = json.loads(res.text)
        data = clean_data(odata)
        if data and "error" in data:
            raise _converter.structure(data, Error)
        return _converter.structure(data, cl)


class Ros(BaseRos):
    _interface: Optional[InterfaceModule] = None
    _ip: Optional[IPModule] = None
    _ppp: Optional[PPPModule] = None
    _queue: Optional[QueueModule] = None
    _routing: Optional[RoutingModule] = None
    _system: Optional[SystemModule] = None
    _tool: Optional[ToolModule] = None
    """
    Ros class that represent a routeros device.

    Attributes
    ----------
    :param str server: The device's Rest API url, example https://192.168.88.1/
    :param str username: Username
    :param str password: Password
    :param requests.Session session: A requests.Session instance
    :param bool secure: Valdiate ssl cert
    :param str filename: Base filename of the rest API, default to `rest`
    :param str url: The device's Rest API url with base filename, default to https://server/rest
    """

    @property
    def bridge(self) -> BridgeModule:
        return self.interface.bridge

    @property
    def interface(self):
        if not self._interface:
            self._interface = InterfaceModule(self)
        return self._interface

    @property
    def ip(self):
        if not self._ip:
            self._ip = IPModule(self)
        return self._ip

    @property
    def ppp(self):
        if not self._ppp:
            self._ppp = PPPModule(self)
        return self._ppp

    @property
    def queue(self):
        if not self._queue:
            self._queue = QueueModule(self)
        return self._queue

    @property
    def routing(self):
        if not self._routing:
            self._routing = RoutingModule(self)
        return self._routing

    @property
    def system(self):
        if not self._system:
            self._system = SystemModule(self)
        return self._system

    @property
    def tool(self):
        if not self._tool:
            self._tool = ToolModule(self, "/tool")
        return self._tool

    def log(self, **kwds: Any):
        return self.get_as("/log", List[Log], kwds)

    def ping(self, address: str, count: int = 4):
        return self.tool.ping(address, count)



from attr import define
from typing import Optional

from ros._base import BaseModule, BaseProps

from .connection import IPConnection


@define
class IPFirewallModule:
    mod: BaseModule
    url: str
    _connection: Optional[BaseProps[IPConnection]] = None

    @property
    def connection(self) -> BaseProps[IPConnection]:
        if not self._connection:
            self._connection = BaseProps(
                self.mod.ros, self.url + "/connection", IPConnection
            )
        return self._connection



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ros',
 'ros.inteface',
 'ros.inteface.bridge',
 'ros.inteface.list',
 'ros.ip',
 'ros.ip.dhcp_server',
 'ros.ip.dns',
 'ros.ip.firewall',
 'ros.ppp',
 'ros.queue',
 'ros.routing',
 'ros.system',
 'ros.system.package',
 'ros.tool']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.0.0,<=23.0.0', 'cattrs>=21.0.0,<=23.0.0', 'requests>=2.28.1,<3.0.0']

extras_require = \
{'ujson': ['ujson>=5.5.0,<6.0.0']}

setup_kwargs = {
    'name': 'rosrestpy',
    'version': '0.4.0',
    'description': 'RouterOS v7 REST API python module',
    'long_description': '# RosRestPy\n\n[![PyPi Package Version](https://img.shields.io/pypi/v/rosrestpy)](https://pypi.org/project/rosrestpy/)\n[![Supported Python versions](https://img.shields.io/pypi/pyversions/rosrestpy)](https://pypi.org/project/rosrestpy/)\n[![LICENSE](https://img.shields.io/github/license/hexatester/rosrestpy)](https://github.com/hexatester/rosrestpy/blob/main/LICENSE)\n\nRouterOS v7 REST API python module\n\n## RouterOS v7 REST API Support\n\nNot all types and methods of the RouterOS v7 REST API are supported, yet.\n\n## Installing\n\nYou can install or upgrade rosrestpy with:\n\n```bash\npip install rosrestpy --upgrade\n```\n\n## Example\n\n```python\nfrom ros import Ros\n\nros = Ros("https://192.168.88.1/", "admin", "")\nif ros.system.resource.cpu_load > 90:\n    print(f"{ros.system.identity}\'s CPU > 90%")\n\nfor interface in ros.interface():\n    print(interface.name)\n\nqueues = ros.queue.simple(name="Hotspot")\nif len(queues) == 1:\n    queue = queues[0]\n    print(f"Usage {queue.bytes}")\n\nbw_tests = ros.tool.bandwith_test("172.16.0.1", "3s", "admin", direction="both")\nresult_bw_test = bw_tests[-1]\nprint(f"Download {result_bw_test.rx_total_average}")\nprint(f"Upload {result_bw_test.tx_total_average}")\n```\n\n## Resources\n\nThe [RouterOS REST API](https://help.mikrotik.com/docs/display/ROS/REST+API) is the technical reference for `rosrestpy`.\n\n## Contributing\n\nContributions of all sizes are welcome. Please review our [contribution guidelines](https://github.com/hexatester/rosrestpy/blob/main/CONTRIBUTING.md "How To Contribute") to get started. You can also help by [reporting bugs or feature requests](https://github.com/hexatester/rosrestpy/issues/new/choose).\n',
    'author': 'hexatester',
    'author_email': 'hexatester@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/hexatester/rosrestpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List

from ros._base import BaseProps
from .dhcp_server import DHCPServer
from .network import DHCPNetwork


class DHCPServerModule(BaseProps[DHCPServer]):
    @property
    def network(self) -> List[DHCPNetwork]:
        return self.ros.get_as(self.filename + "/network", List[DHCPNetwork])


__all__ = ["DHCPNetwork", "DHCPServerModule", "DHCPServer"]



from attr import dataclass
from typing import Optional


@dataclass
class DHCPNetwork:
    address: str
    caps_manager: str
    dhcp_option: str
    dns_server: str
    dynamic: bool
    gateway: str
    ntp_server: str
    wins_server: str
    dns_none: bool = False
    dhcp_option_set: Optional[str] = None
    domain: Optional[str] = None
    next_server: Optional[str] = None
    boot_file_name: Optional[str] = None
    netmask: Optional[int] = None
    comment: Optional[str] = None
    id: str = None



from ros._base import BaseModule, BaseProps

from .interface import InterfaceQueue
from .simple import SimpleQueue


class QueueModule(BaseModule):
    _interface: BaseProps[InterfaceQueue] = None
    _simple: BaseProps[SimpleQueue] = None

    @property
    def interface(self):
        if not self._interface:
            self._interface = BaseProps(self, self.url + "/interface", InterfaceQueue)
        return self._interface

    @property
    def simple(self):
        if not self._simple:
            self._simple = BaseProps(self, self.url + "/simple", SimpleQueue)
        return self._simple



from attr import dataclass
from typing import Optional


@dataclass
class SimpleQueue:
    id: str
    bucket_size: str
    burst_limit: str
    burst_threshold: str
    burst_time: str
    bytes: str
    disabled: bool
    dropped: str
    dynamic: bool
    invalid: bool
    limit_at: str
    max_limit: str
    name: str
    packet_marks: str
    packet_rate: str
    packets: str
    parent: Optional[str]
    priority: str
    queue: str
    queued_bytes: str
    queued_packets: str
    rate: str
    target: str
    total_bytes: int
    total_dropped: int
    total_packet_rate: int
    total_packets: int
    total_queued_bytes: int
    total_queued_packets: int
    total_rate: int
    comment: Optional[str] = None
    dst: Optional[str] = None
    total_queue: Optional[str] = None

    def __str__(self) -> str:
        return self.name

    def __bool__(self) -> bool:
        return not self.disabled



from attr import dataclass
from typing import Optional


@dataclass
class InterfaceEthernet:
    advertise: str
    arp: str
    arp_timeout: str
    auto_negotiation: bool
    bandwidth: str
    default_name: str
    disabled: bool
    driver_rx_byte: int
    driver_rx_packet: int
    driver_tx_byte: int
    driver_tx_packet: int
    full_duplex: bool
    l2mtu: int
    loop_protect: str
    loop_protect_disable_time: str
    loop_protect_send_interval: str
    loop_protect_status: str
    mac_address: str
    mtu: int
    name: str
    orig_mac_address: str
    running: bool
    rx_broadcast: int
    rx_bytes: int
    rx_fcs_error: int
    rx_flow_control: str
    rx_fragment: int
    rx_jabber: int
    rx_multicast: int
    rx_pause: int
    speed: str
    tx_broadcast: int
    tx_bytes: int
    tx_flow_control: str
    tx_multicast: int
    tx_pause: int
    id: str
    comment: Optional[str] = None
    fec_mode: Optional[str] = None
    rx_1024_1518: Optional[int] = None
    rx_128_255: Optional[int] = None
    rx_1519_max: Optional[int] = None
    rx_256_511: Optional[int] = None
    rx_512_1023: Optional[int] = None
    rx_64: Optional[int] = None
    rx_65_127: Optional[int] = None
    rx_align_error: Optional[int] = None
    rx_drop: Optional[int] = None
    rx_length_error: Optional[int] = None
    rx_packet: Optional[int] = None
    rx_too_long: Optional[int] = None
    rx_too_short: Optional[int] = None
    sfp_rate_select: Optional[str] = None
    slave: Optional[bool] = None
    switch: Optional[str] = None
    tx_1024_1518: Optional[int] = None
    tx_128_255: Optional[int] = None
    tx_256_511: Optional[int] = None
    tx_512_1023: Optional[int] = None
    tx_64: Optional[int] = None
    tx_65_127: Optional[int] = None
    tx_collision: Optional[int] = None
    tx_deferred: Optional[int] = None
    tx_drop: Optional[int] = None
    tx_excessive_collision: Optional[int] = None
    tx_fcs_error: Optional[int] = None
    tx_late_collision: Optional[int] = None
    tx_multiple_collision: Optional[int] = None
    tx_packet: Optional[int] = None
    tx_single_collision: Optional[int] = None

    def __str__(self) -> str:
        return self.name
